/*
	This is a ACLfs client.

	It maps various filesystems into the client home/aclfs directory.
	It deals with the aclfsd server to gain access to files. The later
	implement (may implement) sophisticated access control using
	ACL (Access Control List) instead of the limited Unix 9 bits
	permission scheme.
*/
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <unistd.h>
#include <fcntl.h>
#include <wait.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <dirent.h>
#include <limits.h>
#include <unistd.h>
#include <sys/uio.h>
#include <sys/time.h>
#include <utime.h>
#include <signal.h>
#include <sys/stat.h>
#include <sys/fcntl.h>
#include <sys/socket.h>
#include <sys/un.h>
#include "aclfs.h"
#include "../remaudio.h"
#include "../dispatch.h"

#ifndef SCM_CREDENTIALS
	#define SCM_CREDENTIALS 2
	// Struct not available in headers yet, but supported by the kernel
	struct ACL_UCRED {
		int32_t	pid;
		int32_t	uid;
		int32_t	gid;
	};
#else
	typedef struct ucred ACL_UCRED;
#endif

extern "C" char *stpcpy (char *, const char *);

struct ACLFS_CONF {
	char *volume;
	char *realvolume;
};

static int nbdef=0;
static ACLFS_CONF *conf;

static void aclfs_readconf ()
{
	if (conf == NULL){
		/* #Specification: aclfs / global and user config
			The aclfs configuration for a user is the union of
			/etc/aclfs.conf and the user's .virtualfs/aclfs.conf file. If one
			aclfs volume exist in both /etc/aclfs.conf and the
			.aclfs file, the user definition has priority.

			Note that this behavior is weak. We must implement
			something better that merge all entries in the "n"
			directory. A user may have a logical volume in
			litefs that conflict with /etc/aclfs.conf, and the
			definition in .litefs must have priority.

			This will be solve one day once we have a virtualfs
			agent that loads all config file and make the configuration
			available through a shared memory mapping. This utility
			will be able to handle the .persmount, .litefs and .aclfs
			(and what's next :-) ) properly.
		*/
		conf = (ACLFS_CONF*)malloc (100*sizeof(ACLFS_CONF));
		FILE *fin = fopen ("/etc/virtualfs/aclfs.conf","r");
		if (fin != NULL){
			char buf[1000];
			while (fgets(buf,sizeof(buf)-1,fin)!=NULL){
				char *pt = buf;
				while (isspace (*pt)) pt++;
				if (pt[0] != '\0' && pt[0] != '#'){
					char *volume = pt;
					while (*pt > ' ') pt++;
					if (isspace(*pt)) *pt = '\0';
					conf[nbdef].volume = strdup(volume);
					conf[nbdef].realvolume = strdup(volume);
					nbdef++;
				}
			}
			fclose (fin);
		}
		char path[PATH_MAX];
		snprintf (path,sizeof(path)-1,"%s/.virtualfs/aclfs.conf",_dispatch_gethome());
		fin = fopen (path,"r");
		if (fin != NULL){
			char buf[1000];
			while (fgets(buf,sizeof(buf)-1,fin)!=NULL){
				char *pt = buf;
				while (isspace (*pt)) pt++;
				if (pt[0] != '\0' && pt[0] != '#'){
					char *volume = pt;
					while (*pt > ' ') pt++;
					if (isspace(*pt)){
						*pt ++ = '\0';
						while (isspace(*pt)) pt++;
						char *end = pt + strlen(pt)-1;
						if (*end == '\n') *end = '\0';
						int i;
						for (i=0; i<nbdef; i++){
							if (strcmp(conf[i].volume,volume)==0){
								free (conf[i].realvolume);
								conf[i].realvolume = strdup(pt);
								break;
							}
						}
						if (i==nbdef){
							conf[nbdef].volume = strdup(volume);
							conf[nbdef].realvolume = strdup(pt);
							nbdef++;
						}
					}
				}
			}
			fclose (fin);
		}
	}
}

struct ACLFS_INFO{
	char volume[PATH_MAX];		// Sub-directory in $HOME/n
	char relpath[PATH_MAX];		// relative path in $HOME/n/volume/
	char virtpath[PATH_MAX];	// absolute virtual path
	char realpath[PATH_MAX];	// Translated path
};

/*
	Check if a volume is configured for the personal auto mounter
*/
static int aclfs_chkconf (ACLFS_INFO &info)
{
	aclfs_readconf();
	int ret = -1;
	for (int i=0; i<nbdef; i++){
		if (strcmp(info.volume,conf[i].volume)==0){
			sprintf (info.realpath,"/%s%s",conf[i].realvolume
				,info.relpath);
			ret = 0;
			break;
		}
	}
	return ret;
}


static bool aclfs_root = false;
static bool aclfs_root_checked = false;

static bool aclfs_isroot()
{
	if (!aclfs_root_checked){
		aclfs_root = getenv("ACLFS_ROOT")!=NULL;
		aclfs_root_checked = true;
	}
	return aclfs_root;
}

static int aclfs_connect();

/*
	Translate a logical path if it is part of the personnal mount area
	($HOME/n)
*/
static int aclfs_translate (
	const char *path,
	ACLFS_INFO &info,
	ACL_REQUEST &req,
	int &len_request)
{
	// Avoid calling aclfs_isroot() for performance, code is duplicate here
	if (!aclfs_root_checked){
		aclfs_root = getenv("ACLFS_ROOT")!=NULL;
		aclfs_root_checked = true;
	}
	int ret = -1;
	if (aclfs_root){
		ret = 0;
		if (strstr(path,"..")!=NULL){
			_dispatch_simplifypath (info.virtpath,info.virtpath);
			path = info.virtpath;
		}
		strcpy (info.realpath,path);
		if (path[0] != '/'){
			const char *curpath = _dispatch_getcwd();
			snprintf (info.virtpath,sizeof(info.virtpath)-1,"%s/%s",curpath,path);
			path = info.virtpath;
		}else{
			strcpy (info.virtpath,path);
		}
		strcpy (req.paths,"/root");
		char *end = stpcpy (req.paths+5,path) + 1;
		len_request = (int)(end-(char*)&req);
		req.mode = 0;
		req.perm = 0;
	}else{
		int lenn;
		const char *pathn = _dispatch_getpathn(lenn);
		if (pathn == NULL) return -1;
		strcpy (info.realpath,path);
		if (path[0] != '/'){
			const char *curpath = _dispatch_getcwd();
			snprintf (info.virtpath,sizeof(info.virtpath)-1,"%s/%s",curpath,path);
			path = info.virtpath;
		}else{
			strcpy (info.virtpath,path);
		}
		if (strstr(path,"..")!=NULL){
			_dispatch_simplifypath (info.virtpath,info.virtpath);
			path = info.virtpath;
		}

		//printf ("pathn = :%s: <> :%s:\n",pathn,path);
		if (strncmp(pathn,path,lenn)==0){
			path += lenn;
			char *volume = info.volume;
			while (*path != '\0' && *path != '/'){
				*volume++ = *path++;
			}
			*volume = '\0';
			strcpy (info.relpath,path);
			//printf ("volume :%s: :%s:\n",info.volume,info.relpath);
			if (aclfs_chkconf (info)!=-1){
				if (aclfs_connect() != -1){
					char *end = stpcpy (req.paths,info.realpath) + 1;
					len_request = (int)(end-(char*)&req);
					req.mode = 0;
					req.perm = 0;
					ret = 0;
				}
			}
		}
	}
	return ret;
}

static int aclfs_mkdiranc (const char *path)
{
	int ret = __mkdir (path,0755);
	if (ret == -1){
		char tmp[strlen(path)+1];
		strcpy (tmp,path);
		char *pt = strrchr(tmp,'/');
		if (pt != NULL){
			*pt = '\0';
			if (aclfs_mkdiranc (tmp) != -1) ret = __mkdir (path,0755);
		}
	}
	return ret;
}

static void aclfs_fakechdir(const char *path)
{
	if (__chdir (path) == -1){
		aclfs_mkdiranc (path);
		__chdir (path);
	}
}


#define ACLFS_STATIC static

#include "aclfs_common.ch"

static int aclfs_setxattr (const char *path, const char *name,
		     const void *value, size_t size, int flags)
{
	int ret = DISPATCH_DONTCARE;
	ACLFS_INFO info;
	ACL_REQUEST req;
	int len;
	if (aclfs_translate (path,info,req,len)!=-1){
		errno = ENOSYS;
		ret = -1;
	}
	return ret;
}

static int aclfs_lsetxattr (const char *path, const char *name,
		      const void *value, size_t size, int flags)
{
	int ret = DISPATCH_DONTCARE;
	ACLFS_INFO info;
	ACL_REQUEST req;
	int len;
	if (aclfs_translate (path,info,req,len)!=-1){
		errno = ENOSYS;
		ret = -1;
	}
	return ret;
}

static int aclfs_fsetxattr (int fd, const char *name,
		      const void *value, size_t size, int flags)
{
	errno = ENOSYS;
	return -1;
}

static ssize_t aclfs_getxattr (const char *path, const char *name,
			 void *value, size_t size)
{
	ssize_t ret = DISPATCH_DONTCARE;
	ACLFS_INFO info;
	ACL_REQUEST req;
	int len;
	if (aclfs_translate (path,info,req,len)!=-1){
		errno = ENOSYS;
		ret = -1;
	}
	return ret;
}

static ssize_t aclfs_lgetxattr (const char *path, const char *name,
			  void *value, size_t size)
{
	ssize_t ret = DISPATCH_DONTCARE;
	ACLFS_INFO info;
	ACL_REQUEST req;
	int len;
	if (aclfs_translate (path,info,req,len)!=-1){
		errno = ENOSYS;
		ret = -1;
	}
	return ret;
}

static ssize_t aclfs_fgetxattr (int fd, const char *name,
			  void *value, size_t size)
{
	errno = ENOSYS;
	return -1;
}


static ssize_t aclfs_listxattr (const char *path, char *list, size_t size)
{
	ssize_t ret = DISPATCH_DONTCARE;
	ACLFS_INFO info;
	ACL_REQUEST req;
	int len;
	if (aclfs_translate (path,info,req,len)!=-1){
		errno = ENOSYS;
		ret = -1;
	}
	return ret;
}
static ssize_t aclfs_llistxattr (const char *path, char *list, size_t size)
{
	ssize_t ret = DISPATCH_DONTCARE;
	ACLFS_INFO info;
	ACL_REQUEST req;
	int len;
	if (aclfs_translate (path,info,req,len)!=-1){
		errno = ENOSYS;
		ret = -1;
	}
	return ret;
}

static ssize_t aclfs_flistxattr (int fd, char *list, size_t size)
{
	errno = ENOSYS;
	return -1;
}

static int aclfs_removexattr (const char *path, const char *name)
{
	int ret = DISPATCH_DONTCARE;
	ACLFS_INFO info;
	ACL_REQUEST req;
	int len;
	if (aclfs_translate (path,info,req,len)!=-1){
		errno = ENOSYS;
		ret = -1;
	}
	return ret;
}

static int aclfs_lremovexattr (const char *path, const char *name)
{
	int ret = DISPATCH_DONTCARE;
	ACLFS_INFO info;
	ACL_REQUEST req;
	int len;
	if (aclfs_translate (path,info,req,len)!=-1){
		errno = ENOSYS;
		ret = -1;
	}
	return ret;
}
static int aclfs_fremovexattr (int, const char *name)
{
	errno = ENOSYS;
	return -1;
}



static DISPATCH m={
	NULL,
	aclfs_open,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	aclfs_stat,
	NULL,
	aclfs_lstat,
	aclfs_chdir,
	aclfs_opendir,
	aclfs_readdir,
	aclfs_closedir,
	aclfs_rename,
	aclfs_unlink,
	aclfs_rmdir,
	aclfs_mkdir,
	NULL,		// socket
	NULL,		// connect
	aclfs_bind,
	aclfs_chmod,
	aclfs_chown,
	aclfs_lchown,
	aclfs_readlink,
	aclfs_utime,
	aclfs_access,
	aclfs_euidaccess,
	aclfs_link,
	aclfs_symlink,
	aclfs_fchmod,
	aclfs_fchown,
	aclfs_execve,
	aclfs_rewinddir,
	aclfs_seekdir,

	setxattr: aclfs_setxattr,
	lsetxattr: aclfs_lsetxattr,
	fsetxattr: aclfs_fsetxattr,

	getxattr: aclfs_getxattr,
	lgetxattr: aclfs_lgetxattr,
	fgetxattr: aclfs_fgetxattr,

	listxattr: aclfs_listxattr,
	llistxattr: aclfs_llistxattr,
	flistxattr: aclfs_flistxattr,

	removexattr: aclfs_removexattr,
	lremovexattr: aclfs_lremovexattr,
	fremovexattr: aclfs_fremovexattr,

	fcntl: NULL,

};

static REGISTER_DISPATCH register_m (m);


