#pragma implementation
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <limits.h>
#include <unistd.h>
#include <dirent.h>
#include <signal.h>
#include <sys/time.h>
#include <utime.h>
#include <sys/uio.h>
#include <sys/fcntl.h>
#include <sys/socket.h>
#include <sys/un.h>
#include "aclfs.h"
#include "libaclfs.h"
#include <vector>

extern "C" char *stpcpy (char *, const char *);

using namespace std;

struct ACLFS_INFO{
	char volume[PATH_MAX];		// Sub-directory in $HOME/n
	char relpath[PATH_MAX];		// relative path in $HOME/n/volume/
	char virtpath[PATH_MAX];	// absolute virtual path
	char realpath[PATH_MAX];	// Translated path
};

#ifndef SCM_CREDENTIALS
	#define SCM_CREDENTIALS 2
	// Struct not available in headers yet, but supported by the kernel
	struct ACL_UCRED {
		int32_t	pid;
		int32_t	uid;
		int32_t	gid;
	};
#else
	typedef struct ucred ACL_UCRED;
#endif

struct LIBACLFS_CONF{
	int len;
	char *dir;
	char *volume;
	LIBACLFS_CONF(const char *_dir, const char *_volume)
	{
		dir = strdup(_dir);
		len = strlen(_dir);
		volume = strdup(_volume);
	}
	LIBACLFS_CONF(const LIBACLFS_CONF &c)
	{
		dir = strdup(c.dir);
		len = strlen(dir);
		volume = strdup(c.volume);
	}
	~LIBACLFS_CONF()
	{
		free (dir);
		free (volume);
	}
};

static vector<LIBACLFS_CONF> confs;


/*
	Simplify a path by stripping .. and .
	Return -1 if path is invalid, 0 if ok
	No validation are made on disk (for directory existence).

	If result is empty, replace by "." or "/" if the path was  absolute.

	Same code as in ../dispatch.cc, for packaging libaclfs independantly
*/
static int libaclfs_simplifypath (const char *path, char *newp)
{
	int ok = 0;
	char *debnewp = newp;
	char *lastvalid[20];	/* Last member other than . and .. */
	int nbvalid = 0;
	const bool isabs = path[0] == '/';
	char tmp[PATH_MAX];
	if (path == newp){
		strcpy (tmp,path);
		path = tmp;
	}
	while (path[0] == '.' && path[1] == '/') path += 2;
	if (path[0] != '/' && path[0] != '.'){
		lastvalid[nbvalid++] = newp;
	}
	while (ok == 0 && *path	!= '\0'){
		if (path[0] == '/'
			&& path[1] == '.'
			&& (path[2] == '/' || path[2] == '\0')){
			path += 2;
		}else if (path[0] == '/'
			&& path[1] == '.'
			&& path[2] == '.'
			&& (path[3] == '/' || path[3] == '\0')){
			if (nbvalid > 0){
				newp = lastvalid[--nbvalid];
				/* x/../y -> y and not /y */
				if (newp == debnewp
					&& !isabs
					&& path[3] == '/') path++;
			}else if (isabs){
				ok = -1;
				break;
			}else{
				*newp++ = '/';
				*newp++ = '.';
				*newp++ = '.';
				*newp = '\0';
			}
			path += 3;
		}else{
			if (*path == '/') lastvalid[nbvalid++] = newp;
			*newp++ = *path++;
		}
	}
	*newp = '\0';
	if (debnewp[0] == '\0'){
		debnewp[0] = isabs ? '/' : '.';
		debnewp[1] = '\0';
	}
	return ok;
}


static int aclfs_connect ();

static int aclfs_translate (
	const char *path,
	ACLFS_INFO &info,
	ACL_REQUEST &req,
	int &len_request)
{
	int ret =  -1;
	strcpy (info.realpath,path);
	if (path[0] != '/'){
		char cwd[PATH_MAX];
		const char *curpath = getcwd(cwd,PATH_MAX);
		snprintf (info.virtpath,sizeof(info.virtpath)-1,"%s/%s",curpath,path);
		path = info.virtpath;
	}else{
		strcpy (info.virtpath,path);
	}
	if (strstr(path,"..")!=NULL){
		libaclfs_simplifypath (info.virtpath,info.virtpath);
		path = info.virtpath;
	}
	for (int i=0; i<confs.size(); i++){
		LIBACLFS_CONF &c = confs[i];
		// fprintf (stderr,"path = :%s: <> :%s:\n",c.dir,path);
		if (strncmp(c.dir,path,c.len)==0
			&& (path[c.len] == '/' || path[c.len] == '\0')){
			strcpy (info.volume,c.volume);
			path += c.len;
			strcpy (info.relpath,path);
			// fprintf (stderr,"volume :%s: :%s:\n",info.volume,info.relpath);
			if (aclfs_connect() != -1){
				// fprintf (stderr,"Connected\n");
				snprintf (info.realpath,sizeof(info.realpath)-1,"/%s%s"
					,c.volume,info.relpath);
				char *end = stpcpy (req.paths,info.realpath) + 1;
				len_request = (int)(end-(char*)&req);
				req.mode = 0;
				req.perm = 0;
				ret = 0;
			}
			break;
		}
	}
	return ret;
}

static int aclfs_fakechdir (const char *)
{
	return 0;
}

static bool aclfs_isroot()
{
	return true;
}


#define DISPATCH_DONTCARE ACLFS_NOTMANAGED

#define ACLFS_STATIC extern "C"
#include "aclfs_common.ch"


ACLFS_EXTERN_C void aclfs_setmount (const char *dir, const char *volume)
{
	confs.push_back(LIBACLFS_CONF(dir,volume));
}


ACLFS_EXTERN_C void aclfs_setsocket (const char *path)
{
	socket_path = strdup(path);
}


// Various functions for completness

ACLFS_EXTERN_C int aclfs_read (int fd, void *buf, int size)
{
	return read (fd,buf,size);
}

ACLFS_EXTERN_C int aclfs_close (int fd)
{
	return close (fd);
}

ACLFS_EXTERN_C FILE *aclfs_fopen (const char *path, const char *opt)
{
	FILE *ret = NULL;
	int flags = 0;
	if (strchr(opt,'r')!=NULL) flags |= O_RDONLY;
	if (strchr(opt,'w')!=NULL) flags |= O_WRONLY;
	if (strchr(opt,'a')!=NULL) flags |= O_APPEND;
	int fd = aclfs_open (path,flags,0644);
	if (fd != -1){
		ret = fdopen (fd,opt);
	}
	return ret;
}

ACLFS_EXTERN_C int aclfs_fclose (FILE *f)
{
	return fclose (f);
}

ACLFS_EXTERN_C int aclfs_setcred (const char *user, const char *passwd)
{
	int ret = -1;
	if (strlen(user)<64
		&& strlen(passwd)<64
		&& aclfs_connect() != -1){
		ACL_REQUEST req;
		req.cmd = ACL_SETCRED;
		req.mode = 0;
		req.perm = 0;
		char *p = stpcpy (req.paths,user);
		p++;
		p = stpcpy (p,passwd);
		p++;
		aclfs_sendreq (req,(int)(p-(char*)&req));
		ret = aclfs_receive();
	}
	return ret;
}

